
/*
 * whirlgif.c
 *
 * Copyright (C) 1995,1996 by Kevin Kadow (kadokev@msg.net)
 * 
 * Based on txtmerge.c
 * Copyright (C) 1990,1991,1992,1993 by Mark Podlipec. 
 * All rights reserved.
 *
 * This software may be freely copied, modified and redistributed
 * without fee provided that this copyright notice is preserved 
 * intact on all copies and modified copies.
 * 
 * There is no warranty or other guarantee of fitness of this software.
 * It is provided solely "as is". The author(s) disclaim(s) all
 * responsibility and liability with respect to this software's usage
 * or its effect upon hardware or computer systems.
 *
 */
 /*
  * Description:
  *
  * This program reads in a sequence of single-image GIF format files and
  * outputs a single multi-image GIF file, suitable for use as an animation.
  *
  * TODO:
  *
  * More options for dealing with the colormap
  *
  * Eventually, I'd like to have this program compare the current image
  * with the previous image and check to see if only a small section of
  * the screen changed from the previous image. Worth a shot.
  */

 /*
  * Rev 2.01 +2 10Dec99 Rick
  *	read option from indirect file
  *	block copy (good-bye LZW!)
  * Rev 2.01 +1  4Dec99 Rick
  *	colormap optimization
  *
  * Rev 2.01    31Aug96 Kevin Kadow
  *	disposal
  * Rev 2.00	05Feb96 Kevin Kadow
  *	transparency, gif comments,
  * Rev 1.10	29Jan96 Kevin Kadow
  *	first release of whirlgif
  *
  * txtmerge:
  * Rev 1.00	23Jul91	Mark Podlipec
  *	creation
  * Rev 1.01	08Jan92	Mark Podlipec
  *     use all colormaps, not just 1st.
  *
  * 
  */
#define DA_REV 2.01

#include <stdio.h>
#include <stdlib.h>
#ifdef _USE_STRINGS_H
#include <strings.h>
#else
#include <string.h>
#endif
#ifdef __HUMAN68K__
#include <fcntl.h>
#include <unistd.h>
#endif

#include "whirlgif.h"

#define MAXVAL  4100            /* maxval of lzw coding size */
#define MAXVALP 4200

typedef struct {
	int	type;
	int	i;
	int	argc;
	char	**argv;
	int	count;
	FILE	*infile;
	FILE	*fout;
	char	*param;
	char	str[BIGSTRING];
} FSET;

#define	OPTION	2
#define	FILENAME	1

/*
 * Set some defaults, these can be changed on the command line
 */
unsigned int	loop = DEFAULT_LOOP, loopcount = 0,
/*	use_colormap = DEFAULT_USE_COLORMAP, */
#ifdef _DEBUG_
	debug_flag = 0,
#endif
	verbose = 0;

int imagex;
int imagey;
int imagec;

/* global settings for offset, transparency */
Global	global;
GIF_Color	gif_cmap[256];

GIF_Screen_Hdr	gifscrn;
GIF_Image_Hdr	gifimage;

ULONG	gif_ptwo[16] = {1,2,4,8,16,32,64,128,256,512,1024,2048,4096,8192,0,0};

int	screen_was_last;


void	TheEnd();
void	TheEnd1(char *p);
void	Usage();
void	init();
char	*getword(char *str);
int	getarg(FSET *f);
char	*getparam(FSET *f);
void	GIF_CopyBlocks(FILE *fp, FILE *fout);
void	GIF_Screen_Header(FILE *fp, FILE *fout, int first_name);
void	GIF_Image_Header(FILE *fp, FILE *fout, int first_name);
void	GIF_Read_File(FILE *fout, char *fname, int first_image);
int	GIF_Get_Short(FILE *fp, FILE *fout, int first_time);
void	GIF_Comment(FILE *fout, char *string);
void	GIF_Loop(FILE *fout, unsigned int repeats);
void	GIF_GCL(FILE *fout, unsigned int delay);
void	Calc_Trans(char *string);
void	set_offset(char *string);


void
TheEnd()
{
	exit(0);
}

void
TheEnd1(p)
char	*p;
{
	fprintf(stderr, p);
	TheEnd();
}

void
Usage()
{
	fprintf(stderr, "\nUsage: whirlgif [-o outfile] [-l [count]] [-time #delay]\n");
	fprintf(stderr, "\t[-disp [none|not|back|prev]] [-trans index] [-off x,y]\n");
	fprintf(stderr, "\t[-i listfile] file1 [ -time #delay] file2 ...\n");
	exit(0);
}

void
init()
{
	/* set global values */
	screen_was_last = FALSE;
	global.trans.type = TRANS_NONE;
	global.trans.valid = FALSE;
	global.time = DEFAULT_TIME;
	global.left = 0;
	global.top = 0;
	global.disposal = DEFAULT_DISPOSAL;

	imagex = 0;
	imagey = 0;
	imagec = 0;
}

int
getarg(f)
FSET	*f;
{
	char	*p;

	if (f->infile) {
		if (fgets(f->str, BIGSTRING, f->infile)) {
			p = f->str;
			if (*p++ == ':') {
				char	*w = f->str;

				while (*p && *p > ' ')	*w++ = *p++;
				*w = '\0';

				while (*p && *p <= ' ')	p++;
				f->param = (*p > ' ')? p:NULL;
				f->type = OPTION;
			} else {
				f->param = NULL;
				f->type = FILENAME;
			}
			while (*p && *p > ' ')	p++;
			if (*p)	*p = '\0';

			return TRUE;
		} else {
			fclose(f->infile);
			f->infile = NULL;
		}
	}
	if (f->i < f->argc) {
		p = f->argv[f->i++];
		if (*p == '-') {
			++p; /* strip off the - */
			strcpy(f->str, p);
			f->type = OPTION;
		} else {
			strcpy(f->str, p);
			f->type = FILENAME;
		}
		return TRUE;
	} else {
		return FALSE;
	}
}

char *
getparam(f)
FSET	*f;
{
	if (f->infile) {
#ifdef _DEBUG_
		fprintf(stderr, "Param: %s\n", f->param);
#endif
		return f->param;
	} else if (f->i < f->argc && f->argv[f->i][0] != '-') {
#ifdef _DEBUG_
		fprintf(stderr, "Param: %s\n", f->argv[f->i]);
#endif
		return f->argv[f->i++];
	} else {
		return NULL;
	}
}

void
main(argc, argv)
int	argc;
char	*argv[];
{
	FSET	f;
	char	temp[BIGSTRING];

	fprintf(stderr, "whirlgif Rev %2.2f (C) 1996 by Kevin Kadow\n", DA_REV);
	fprintf(stderr, "                  (C) 1991,1992 by Mark Podlipec\n");

	if (argc < 2)	Usage();

	f.argc = argc;
	f.argv = argv;
	f.count = 0;
	f.infile = NULL;
	f.fout = stdout;
	f.i = 1;

	init();

#ifdef __HUMAN68K__
	setmode(STDOUT_FILENO, O_BINARY);
#endif

	while (getarg(&f)) {
		if (f.type == OPTION) {
			char	*p;
#ifdef _DEBUG_
			fprintf(stderr, "Option: %s\n", f.str);
#endif
			switch (f.str[0]) {
			case 'v':	/* Give lots of information */
				verbose++;
				fprintf(stderr, "Verbose output\n");
				break;

			case 'd':	/* either Debug mode or disposal setting */
				if (!strncmp("disp", f.str, 4)) {
					if (p = getparam(&f)) {
						if (verbose)	fprintf(stderr, "Disposal method set\n");
						if (!strcmp("none", p))
							global.disposal = DISP_NONE;
						else if (!strcmp("not", p))
							global.disposal = DISP_NOT;
						else if (!strcmp("back", p)) 
							global.disposal = DISP_BACK;
						else if (!strcmp("prev", p)) 
							global.disposal = DISP_PREV;
						else global.disposal = DISP_NONE;
					} else global.disposal = DISP_NONE;
				}
#ifdef _DEBUG_
				else {
					debug_flag++;
					fprintf(stderr, "DEBUG: Debug Level %d\n", debug_flag);
				}
#endif
				break;

			case 'l':	/* Enable looping */
				loop = TRUE;
				if (p = getparam(&f)) {
					/* a loop count was given */
					loopcount = atoi(p);
					if (verbose)	fprintf(stderr, "Loop %d times\n", loopcount);
				} else {
					/* default to infinite loop */
					loopcount = 0;
					if (verbose)	fputs("Looping enabled\n", stderr);
				}
				break;

			case 't':	/* either time or transparent */
				if (!strcmp("time", f.str)) {
					/* Delay time in 1/100's of a second */
					if (p = getparam(&f)) {
						global.time = atoi(p);
					}
				} else if (!strncmp("trans", f.str, 4)) {
					if (p = getparam(&f)) {
						Calc_Trans(p);
					}
				}
				break;

			case 'o':	/* Output file - send output to a given filename */
				if (!strncmp("off", f.str, 3)) {
					if (p = getparam(&f)) {
						set_offset(p);
					}
				} else {
					if (f.count > 0) {
						fprintf(stderr, "Processed %d files.\n", f.count);

						fputc(';', f.fout); /* image separator */
						sprintf(temp, "whirlgif %2.2f (C) kadokev@msg.net. %d images",
									DA_REV, f.count);
						GIF_Comment(f.fout, temp);
						f.count = 0;
					}
					fclose(f.fout);

					if (p = getparam(&f)) {
					/* It must be 'output, so do that */
#ifndef __HUMAN68K__
						if (NULL == (f.fout = fopen(p, "w")))
#else
						if (NULL == (f.fout = fopen(p, "wb")))
#endif
						{
							fprintf(stderr, "Cannot open %s for output\n", p);
							exit(1);
						}
						fprintf(stderr, "Create %s\n", p);
					}
				}
				break;

			case 'i':	/* input file - file with a list of images */
				if (p = getparam(&f)) {
					if (NULL == (f.infile = fopen(p, "r"))) {
						fprintf(stderr, "Cannot read list file %s\n", p);
					}
				}
				break;

			default:
				Usage();
				exit(0);
				break;
			}
			continue;
		}
		/* Not an option, must be the name of an input file */
		GIF_Read_File(f.fout, f.str, !f.count++);
	}
	/* We're done with all the options, finish up */
	if (f.count > 0) {
		fputc(';', f.fout); /* image separator */
		sprintf(temp, "whirlgif %2.2f (C) kadokev@msg.net. %d images", DA_REV, f.count);
		GIF_Comment(f.fout, temp);
	}

	fclose(f.fout);
	fprintf(stderr, "Processed %d files.\n", f.count);
	exit(0);
}


/*
 * Read a GIF file, outputting to fname as we go.
 * It would be faster to read and write the individual blocks,
 * but eventually we'd like to optimize based on changes from
 * previous images(ie only a small section of the image changed.
 */
void
GIF_Read_File(fout, fname, first_image)
FILE	*fout;
char	*fname;
int	first_image;
{
	FILE	*fp;
	int	i;

#ifndef __HUMAN68K__
	if ((fp = fopen(fname, "r")) == 0)
#else
	if ((fp = fopen(fname,"rb")) == 0)
#endif
	{
		fprintf(stderr, "Can't open %s for reading.\n", fname); 
		TheEnd();
	}

	GIF_Screen_Header(fp, fout, first_image);

	/*** read until  ,  separator */
	do {
		i = fgetc(fp);
		if ((i < 0) && feof(fp)) {
			fclose(fp);
			TheEnd1("GIF_Read_Header: Unexpected End of File\n");
		}
	} while(i != ',');

	if (first_image) {
		/* stuff we only do once */
		if (loop)	GIF_Loop(fout, loopcount);
	}
	if (global.time || (global.trans.type != TRANS_NONE && global.trans.valid))
	GIF_GCL(fout, global. time);

	fputc(',', fout); /* image separator */

	GIF_Image_Header(fp, fout, first_image);

	/*FOO*/

	/*** Setup ACTION for IMAGE */

	GIF_CopyBlocks(fp, fout);
	fputc(0, fout);  /* block count of zero */

	fclose(fp);
}


void
GIF_CopyBlocks(fp, fout)
FILE	*fp, *fout;
{
	int	gif_block_size;
	register int	tmp;

	/* root_code_size */
	tmp = (fgetc(fp) & 0xff); fputc(tmp, fout);

	while ((tmp = fgetc(fp)) > 0) {
		fputc(tmp, fout);
		gif_block_size = tmp;
#ifdef _DEBUG_
		fprintf(stderr, "block size: %d\n", gif_block_size);
#endif
		while (--gif_block_size >= 0) {
			tmp = fgetc(fp);
			if (tmp >= 0) {
				fputc(tmp, fout);
			} else {
				TheEnd1("EOF in data stream\n");
			}
		}
	}
	if (tmp < 0)	TheEnd1("EOF in data stream\n");
}


/* 
 * read GIF header 
 */
void
GIF_Screen_Header(fp, fout, first_time)
FILE	*fp, *fout;
int	first_time;
{
	int	temp,i;

	for(i = 0; i < 6; i++) {
		temp = fgetc(fp);
		if (i == 4 && temp == '7')	temp = '9';
		if (first_time == TRUE)	fputc(temp, fout);
	}

	gifscrn.width  = GIF_Get_Short(fp, fout, first_time);
	gifscrn.height = GIF_Get_Short(fp, fout, first_time);
	temp = fgetc(fp);	if (first_time == TRUE)	fputc(temp, fout);
	gifscrn.m       =  temp & 0x80;
	gifscrn.cres    = (temp & 0x70) >> 4;
	gifscrn.pixbits =  temp & 0x07;

	gifscrn.bc  = fgetc(fp);
	if (first_time == TRUE) {
		/* we really should set the background color to the transparent color */
		fputc(gifscrn.bc, fout);
	}

	temp = fgetc(fp);	if (first_time == TRUE)	fputc(temp, fout);
	imagec = gif_ptwo[(1+gifscrn.pixbits)];

	if (verbose)
		fprintf(stderr, "Screen: %dx%dx%d m=%d cres=%d bkgnd=%d pix=%d\n",
				gifscrn.width, gifscrn.height, imagec, gifscrn.m, gifscrn.cres,
				gifscrn.bc, gifscrn.pixbits);

	screen_was_last = first_time;
	if (global.trans.type == TRANS_RGB)	global.trans.valid = 0;
	if (gifscrn.m) {
		for (i = 0; i < imagec; i++) {
			int	red, green, blue;
			red   = fgetc(fp);
			green = fgetc(fp);
			blue  = fgetc(fp);

			if (screen_was_last) {
				if (first_time == TRUE) {
					fputc(red, fout);
					fputc(green, fout);
					fputc(blue, fout);
				}
				gif_cmap[i].cmap.red   = red;
				gif_cmap[i].cmap.green = green;
				gif_cmap[i].cmap.blue  = blue;
			} else {
				if (gif_cmap[i].cmap.red != red
						|| gif_cmap[i].cmap.green != green
						|| gif_cmap[i].cmap.blue != blue) {
					gif_cmap[i].cmap.red   = red;
					gif_cmap[i].cmap.green = green;
					gif_cmap[i].cmap.blue  = blue;
					screen_was_last = TRUE;
				}
			}
			if(global.trans.type == TRANS_RGB && !global.trans.valid)
				if(global.trans.red == gif_cmap[i].cmap.red &&
						global.trans.green == gif_cmap[i].cmap.green &&
						global.trans.blue == gif_cmap[i].cmap.blue) {
#ifdef _DEBUG_
					if (debug_flag > 1)
						fprintf(stderr, " Transparent match at %d\n", i);
#endif
					global.trans.map = i;
					global.trans.valid = TRUE;
				}
		}
	}
}

void
GIF_Image_Header(fp, fout, first_time)
FILE	*fp, *fout;
int	first_time;
{
	int	temp, tnum, i;

	gifimage.left   = GIF_Get_Short(fp, fout, 1);
	if (global.left)	gifimage.left += global.left;

	gifimage.top    = GIF_Get_Short(fp, fout, 1);
	if (global.top)	gifimage.top += global.top;

	gifimage.width  = GIF_Get_Short(fp, fout, 1);
	gifimage.height = GIF_Get_Short(fp, fout, 1);
	temp = fgetc(fp);



	gifimage.i        = temp & 0x40;
	gifimage.pixbits  = temp & 0x07;
	gifimage.m        = temp & 0x80;

	/* this sets the local colormap bit to true */
	if (screen_was_last && (first_time == FALSE))	temp |= 0x80;

	temp &= 0xf8;
	temp |= gifscrn.pixbits;
	fputc(temp, fout);

	imagex = gifimage.width;
	imagey = gifimage.height;
	tnum = gif_ptwo[(1 + gifimage.pixbits)];
	if (verbose)
		fprintf(stderr, "Image: %dx%dx%d (%d,%d) m=%d i=%d pix=%d \n",
				imagex, imagey, tnum, gifimage.left, gifimage.top,
				gifimage.m, gifimage.i, gifimage.pixbits);

	/* if there is an image cmap, then use it */

	if (gifimage.m) {
#ifdef _DEBUG_
		if (debug_flag)	fprintf(stderr, "DEBUG:Transferring colormap of %d colors\n",
									imagec);
#endif
		for (i = 0; i < tnum; i++) {
			fputc(fgetc(fp), fout);
			fputc(fgetc(fp), fout);
			fputc(fgetc(fp), fout);
		}
	}  /* else if screen was last not 1st time */
	else if (screen_was_last && (first_time == FALSE))
	{
#ifdef _DEBUG_
		if (debug_flag > 1)	fprintf(stderr, "DEBUG:Writing colormap of %d colors\n",
								imagec);
#endif
		for (i = 0; i < imagec; i++) {
			fputc(gif_cmap[i].cmap.red  , fout);
			fputc(gif_cmap[i].cmap.green, fout);
			fputc(gif_cmap[i].cmap.blue , fout);
		}
	}
	screen_was_last = FALSE;
}


/*
 *
 */
int
GIF_Get_Short(fp, fout, first_time)
FILE	*fp,*fout;
int	first_time;
{
	register int	temp, tmp1;

	temp = fgetc(fp);	if (first_time == TRUE)	fputc(temp, fout);
	tmp1 = fgetc(fp);	if (first_time == TRUE)	fputc(tmp1, fout);
	return (temp | ((tmp1) << 8));
}

void
GIF_Comment(fout, string)
FILE	*fout;
char	*string;
{
	if (!string || !strlen(string)) {
		/* Bogus string */
#ifdef _DEBUG_
		if (debug_flag)	fprintf(stderr, "GIF_Comment: invalid argument");
#endif
		return;
	}
	fputc(0x21, fout);
	fputc(0xFE, fout);
	fputs(string, fout);
	fputc(0, fout);
}

/*
 * Write a Netscape loop marker.
 */
void
GIF_Loop(fout, repeats)
FILE	*fout;
unsigned int	repeats;
{
	UBYTE	low = 0, high = 0;

	if (repeats) {
		/* non-zero repeat count- Netscape hasn't implemented this yet */
		high = repeats / 256;
		low = repeats % 256;
	}

	fputc(0x21, fout);
	fputc(0xFF, fout);
	fputc(0x0B, fout);
	fputs("NETSCAPE2.0", fout);
	fputc(0x03, fout);
	fputc(0x01, fout);

	fputc(low, fout); /* the delay count - 0 for infinite */
	fputc(high, fout); /* second byte of delay count */
	fputc(0x00, fout); /* terminator */

	if (verbose)	fprintf(stderr, "Wrote loop extension\n");
}

/*
 * GIF_GCL - add a Control Label to set an inter-frame delay value.
 */
void
GIF_GCL(fout, delay)
FILE	*fout;
unsigned int	delay;
{
	UBYTE	low = 0, high = 0, flag;
	if (delay) {
		/* non-zero delay, figure out low/high bytes */
		high = delay / 256;
		low = delay % 256;
	}

	fputc(0x21, fout);
	fputc(0xF9, fout);
	fputc(0x04, fout);

	flag = global.disposal << 2;
	if (delay)	flag |= 0x80;
	if (global.trans.valid)	flag |= 0x01;
	fputc(flag, fout);

	fputc(low, fout); /* the delay speed - 0 is instantaneous */
	fputc(high, fout); /* second byte of delay count */

	fputc(global. trans.map, fout);
	fputc(0, fout);
#ifdef _DEBUG_
	if (debug_flag > 1) {
		fprintf(stderr, "GCL: delay %d", delay);
		if (global.trans. valid)	fprintf(stderr, " Transparent: %d", global. trans.map);
		fputc('\n', stderr);
	}
#endif
}


void
Calc_Trans(string)
char	*string;
{
	if (string[0] != '#') {
		global.trans.type = TRANS_MAP;
		global.trans.map = atoi(string);
		global.trans.valid = 1;
	} else {
		/* it's an RGB value */
		int	r,g,b;

		string++;
		if (3 == sscanf(string, "%2x%2x%2x", &r, &g, &b)) {
			global.trans.red = r;
			global.trans.green = g;
			global.trans.blue = b;
			global.trans.type = TRANS_RGB;
#ifdef _DEBUG_
			if (debug_flag)	fprintf(stderr, "Transparent RGB=(%x,%x,%x)\n", r, g, b);
#endif
		}
	}
#ifdef _DEBUG_
	if (debug_flag)	fprintf(stderr, "DEBUG:Calc_trans is %d\n", global. trans. type);
#endif
}

void
set_offset(string)
char	*string;
{
	char	*off_x,*off_y;

	off_x = (char*)strtok(string, ",");
	off_y = (char*)strtok((char*)NULL, ",");
	if(off_x && off_y) {
		/* set the offset */
		global.left = atoi(off_x);
		global.top = atoi(off_y);
#ifdef _DEBUG_
		if (debug_flag > 1)	fprintf(stderr, "Offset changed to %d,%d\n",
								global. left, global.top);
#endif
		return;
	}
#ifdef _DEBUG_
	if (debug_flag > 1)	fprintf(stderr, "Couldn't parse offset values.\n");
#endif
	exit(1);
}
